# coding: utf-8
"""
Django-rest-easy provides base classes for API views and serializers.

To leverage the QOL features of django-rest-easy, you should use the followint base classes for your serializers:

* :class:`rest_easy.serializers.Serializer`
* :class:`rest_easy.serializers.ModelSerializer`

And if it's model-based, it should use one of the base views provided in the :mod:`rest_easy.views`
module - preferably :class:`rest_easy.views.ModelViewSet` or :class:`rest_easy.views.ReadOnlyModelViewSet`.

As a baseline, all responses using django-rest-easy extension will contain top-level model and schema fields.

Guidelines regarding schemas are as usual: they have to be 100% backwards compatible. In the case of breaking changes,
a serializer with new schema should be created, and the old one slowly faded away - and removed only when no
applications use it - or when it's decided that the feature can't be supported anymore.

An alternative to multi-version fadeout is single-version fadeout, where the change is implemented as a set of
acceptable changes (that is, you can remove the old field only when all clients stop using it - even if it means
sending duplicate data for quite some time).

The classes from this module don't disable any behaviour inherent to Django Rest Framework - anything that is possible
there will be possible with the django-rest-easy base classes.

Django Rest Easy uses following settings:

* REST_EASY_AUTOIMPORT_SERIALIZERS_FROM - for autoimporting serializers.
* REST_EASY_VIEW_BASES - for prepending bases to all views declared in django-rest-easy. They will end up before
  all base views, either DRF's or django-rest-easy's, but after generic mixins in the final generic view mro.
  So in :class:`rest_easy.views.GenericAPIView` and :class:`rest_easy.views.GenericAPIViewSet` they will be at the
  very beginning of the mro, but everything declared in generic mixins, like DRF's CreateMixin, will override that.
* REST_EASY_GENERIC_VIEW_MIXINS - for prepending bases to generic views. They will end up at the beginning of mro
  of all generic views available in django-rest-easy. This can be used to make views add parameters when doing
  perform_update() or perform_create().
* REST_EASY_SERIALIZER_CONFLICT_POLICY - what happens when serializer with same model and schema is redefined. Defaults
  to 'allow', can also be 'raise' - in the former case the new serializer will replace the old one. Allow is used
  to make sure that any import craziness is not creating issues by default.
"""
from django.apps import AppConfig
from django.conf import settings

default_app_config = 'rest_easy.ApiConfig'


class ApiConfig(AppConfig):
    """
    AppConfig autoimporting serializers.

    It scans all installed applications for modules specified in settings.REST_EASY_AUTOIMPORT_SERIALIZERS_FROM
    parameter, trying to import them so that all residing serializers using
    :class:`rest_easy.serializers.SerializerCreator` metaclass will be added to
    :class:`rest_easy.registers.SerializerRegister`.

    In the case of a module not being present in app's context, the import is skipped.
    In the case of a module existing but failing to import, an exception will be raised.
    """
    name = 'rest_easy'
    label = 'rest_easy'

    default_paths = ['serializers', 'api.serializers']

    @property
    def paths(self):
        """
        Get import paths - from settings or defaults.
        """
        return getattr(settings, 'REST_EASY_AUTOIMPORT_SERIALIZERS_FROM', self.default_paths)

    def autodiscover(self):
        """
        Auto-discover serializers in installed apps, fail silently when not present, re-raise exception when present
        and import fails. Borrowed form django.contrib.admin with added nested presence check.
        """

        from importlib import import_module
        from django.apps import apps
        from django.utils.module_loading import module_has_submodule

        for app_config in apps.get_app_configs():
            app = app_config.name
            mod = import_module(app)

            # Attempt to import the app's serializers.
            for item in self.paths:
                try:
                    import_module('{}.{}'.format(app, item))

                except (TypeError, ImportError):
                    # Decide whether to bubble up this error. If the app just
                    # doesn't have serializers module, we can ignore the error
                    # attempting to import it, otherwise we want it to bubble up.
                    curr = mod
                    curr_path = app
                    for part in item.split('.'):  # pragma: no cover
                        if not module_has_submodule(curr, part):
                            break
                        curr_path += '.' + part
                        curr = import_module(curr_path)
                    else:  # pragma: no cover
                        raise

    def ready(self):
        self.autodiscover()



# coding: utf-8
# pylint: skip-file
"""
Tests for django-rest-easy. So far not ported from proprietary code.
"""
from __future__ import unicode_literals

import os
import sys

from django.conf import settings
import django


if __name__ == '__main__':
    settings.configure(DEBUG_PROPAGATE_EXCEPTIONS=True,
                       DATABASES={
                           'default': {
                               'ENGINE': 'django.db.backends.sqlite3',
                               'NAME': ':memory:'
                           }
                       },
                       SITE_ID=1,
                       SECRET_KEY='not very secret in tests',
                       USE_I18N=True,
                       USE_L10N=True,
                       STATIC_URL='/static/',
                       TEMPLATES=[
                           {
                               'BACKEND': 'django.template.backends.django.DjangoTemplates',
                               'APP_DIRS': True,
                           },
                       ],
                       MIDDLEWARE_CLASSES=(
                           'django.middleware.common.CommonMiddleware',
                           'django.contrib.sessions.middleware.SessionMiddleware',
                           'django.contrib.auth.middleware.AuthenticationMiddleware',
                           'django.contrib.messages.middleware.MessageMiddleware',
                       ),
                       INSTALLED_APPS=(
                           'django.contrib.auth',
                           'django.contrib.contenttypes',
                           'django.contrib.sessions',
                           'django.contrib.sites',
                           'django.contrib.staticfiles',
                           'rest_framework',
                           'rest_easy',
                           'rest_easy.tests'
                       ),
                       PASSWORD_HASHERS=(
                           'django.contrib.auth.hashers.MD5PasswordHasher',
                       ),
                       REST_EASY_VIEW_BASES=['rest_easy.tests.mixins.EmptyBase'],
                       REST_EASY_GENERIC_VIEW_MIXINS=['rest_easy.tests.mixins.EmptyMixin'])
    django.setup()

    parent = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, parent)

    from django.test.runner import DiscoverRunner

    runner_class = DiscoverRunner
    test_args = ['tests']

    failures = runner_class(
        verbosity=1, interactive=True, failfast=False).run_tests(test_args)
    sys.exit(failures)



"""
This module contains exceptions native to django-rest-easy.
"""

__all__ = ['RestEasyException']


class RestEasyException(Exception):
    """
    Default django-rest-easy exception class.
    """



# coding: utf-8
"""
This module contains the serializer register.

The serializer register is where all serializers created using
:class:`rest_easy.serializers.SerializerCreator` are registered and where they can be obtained from based
on model and schema. Remember that no other serializers will be kept here - and they will not be obtainable in such
a way.
"""
from __future__ import unicode_literals

import six

from rest_easy.exceptions import RestEasyException
from rest_easy.patterns import BaseRegister

__all__ = ['SerializerRegister', 'serializer_register']


class SerializerRegister(BaseRegister):
    """
    Obtains serializer registration name based on model and schema.
    """
    @staticmethod
    def get_name(model, schema):
        """
        Constructs serializer registration name using model's app label, model name and schema.
        :param model: a Django model, a ct-like app-model string (app_label.modelname) or explicit None.
        :param schema: schema to be used.
        :return: constructed serializer registration name.
        """
        if model is None:
            return schema
        if isinstance(model, six.string_types):
            return '{}.{}'.format(model, schema)
        try:
            return '{}.{}.{}'.format(model._meta.app_label, model._meta.model_name, schema)  # pylint: disable=protected-access
        except AttributeError:
            raise RestEasyException('Model must be either None, a ct-like model string or Django model class.')

    def get(self, model, schema):
        """
        Shortcut to get serializer having model and schema.
        """
        return self.lookup(self.get_name(model, schema))

serializer_register = SerializerRegister()



# coding: utf-8
"""
This module provides useful model mixins and global functions.

Its contents can be used to serialize a model or find proper serializer/deserialize data via a registered serializer.
"""

from __future__ import unicode_literals

from rest_easy.exceptions import RestEasyException
from rest_easy.registers import serializer_register

__all__ = ['SerializableMixin', 'get_serializer', 'deserialize_data']


class SerializableMixin(object):
    """
    This mixin provides serializing functionality to Django models.

    The serializing is achieved thanks to serializers registered in
    :class:`rest_easy.registers.SerializerRegister`. A proper serializer based on model and provided
    schema is obtained from the register and the serialization process is delegated to it.

    Usage:

    ```python
    from users.models import User
    serializer = User.get_serializer(User.default_schema)
    ```
    Or:

    ```python
    data = User.objects.all()[0].serialize()
    ```
    """
    default_schema = 'default'

    @classmethod
    def get_serializer(cls, schema):
        """
        Get correct serializer for this model and given schema,

        Utilizes :class:`rest_easy.registers.SerializerRegister` to obtain correct serializer class.
        :param schema: schema to be used for serialization.
        :return: serializer class.
        """
        name = serializer_register.get_name(cls, schema)
        return serializer_register.lookup(name)

    def serialize(self, schema=None):
        """
        Serialize the model using given or default schema.
        :param schema: schema to be used for serialization or self.default_schema
        :return: serialized data (a dict).
        """
        serializer = self.get_serializer(schema or self.default_schema)
        if not serializer:
            raise RestEasyException('No serializer found for model {} schema {}'.format(self.__class__, schema))
        return serializer(self).data


def get_serializer(data):
    """
    Get correct serializer for dict-like data.

    This introspects model and schema fields of the data and passes them to
    :class:`rest_easy.registers.SerializerRegister`.
    :param data: dict-like object.
    :return: serializer class.
    """
    if 'model' not in data or 'schema' not in data:
        raise RestEasyException('Both model and schema must be provided in data~.')
    serializer = serializer_register.lookup(serializer_register.get_name(data['model'], data['schema']))
    if not serializer:
        raise RestEasyException('No serializer found for model {} schema {}'.format(data['model'], data['schema']))
    return serializer


def deserialize_data(data):
    """
    Deserialize dict-like data.

    This function will obtain correct serializer from :class:`rest_easy.registers.SerializerRegister`
    using :func:`rest_easy.models.get_serializer`.
    :param data: dict-like object or json string.
    :return: Deserialized, validated data.
    """
    serializer = get_serializer(data)(data=data)
    serializer.is_valid(raise_exception=True)
    return serializer.validated_data



# coding: utf-8
"""
This module contains fields necessary for the django-rest-easy module.
"""
from __future__ import unicode_literals

from rest_framework.fields import Field

__all__ = ['StaticField']


class StaticField(Field):  # pylint: disable=abstract-method
    """
    A field that always provides the same value as output.

    The output value is set on initialization, ie::

        from rest_easy.serializers import Serializer

        class MySerializer(Serializer):
            static = StaticField('This will always be the value.')

    """

    def __init__(self, value, **kwargs):
        """
        Initialize the instance with value and DRF settings.
        """
        kwargs['source'] = '*'
        kwargs['read_only'] = True
        super(StaticField, self).__init__(**kwargs)
        self.value = value

    def to_representation(self, value):
        """
        Return the static value.
        """
        return self.value



# coding: utf-8
"""
This module provides scopes usable with django-rest-easy's generic views.

See :mod:`rest_easy.views` for detailed explanation.
"""
from __future__ import unicode_literals

from django.db.models import QuerySet, Model
from django.http import Http404
from django.shortcuts import get_object_or_404

from rest_easy.exceptions import RestEasyException

__all__ = ['ScopeQuerySet', 'UrlKwargScopeQuerySet', 'RequestAttrScopeQuerySet']


class ScopeQuerySet(object):
    """
    This class provides a scope-by-parent-element functionality to views and their querysets.

    It works by selecting a proper parent model instance and filtering view's queryset with it automatically.
    """

    def __init__(self, qs_or_obj, parent_field='pk', related_field=None, raise_404=False, allow_none=False,
                 get_object_handle='', parent=None):
        """
        Sets instance properties, infers sane defaults and ensures qs_or_obj is correct.

        :param qs_or_obj: This can be a queryset or a Django model or explicit None (for particular subclasses)
        :param parent_field: the field to filter by in the parent queryset (qs_or_obj), by default 'id'.
        :param related_field: the field to filter by in the view queryset, by default model_name.
        :param raise_404: whether 404 should be raised if parent object cannot be found.
        :param allow_none: if filtering view queryset by object=None should be allowed. If it's false, resulting
         queryset is guaranteed to be empty if parent object can't be found and 404 is not raised.
        :param get_object_handle: the name under which the object should be available in view. ie.
         view.get_scoped_object(get_object_handle) or view.get_{get_object_handle}. If None, the object will
         not be available from view level. By default will be infered to qs_or_obj's model_name.
        :param parent: if this object's queryset should be filtered by another parameter, parent attribute should be
         an instance of ScopeQuerySet. This allows for ScopeQuerySetChaining (ie. for messages we might have
         UrlKwargScopeQuerySet(User, parent=UrlKwargScopeQuerySet(Account))  for scoping by user and limiting users
         to an account.
        """
        if isinstance(qs_or_obj, QuerySet):
            self.queryset = qs_or_obj
        elif isinstance(qs_or_obj, type) and issubclass(qs_or_obj, Model):
            self.queryset = qs_or_obj.objects.all()
        elif qs_or_obj is None:
            self.queryset = None
        else:
            raise RestEasyException('Queryset parameter must be an instance of QuerySet or a Model subclass.')

        if related_field is None:
            try:
                related_field = '{}'.format(self.queryset.model._meta.model_name)  # pylint: disable=protected-access
            except AttributeError:
                raise RestEasyException('Either related_field or qs_or_obj must be given.')
        self.parent_field = parent_field
        self.related_field = related_field
        self.raise_404 = raise_404
        self.parent = ([parent] if isinstance(parent, ScopeQuerySet) else parent) or []
        self.allow_none = allow_none
        self.get_object_handle = get_object_handle
        if self.get_object_handle == '':
            try:
                self.get_object_handle = self.queryset.model._meta.model_name  # pylint: disable=protected-access
            except AttributeError:
                raise RestEasyException('Either qs_or_obj or explicit get_object_handle (can be None) must be given.')

    def contribute_to_class(self, view):
        """
        Put self.get_object_handle into view's available handles dict to allow easy access to the scope's get_object()
        method in case the object needs to be reused (ie. in child object creation).
        :param view: View the scope is added to.
        """
        if self.get_object_handle:
            if self.get_object_handle in view.rest_easy_available_object_handles:
                raise RestEasyException(
                    'ImproperlyConfigured: multiple scopes with {} get_object handle!'.format(self.get_object_handle)
                )
            view.rest_easy_available_object_handles[self.get_object_handle] = self
        for scope in self.parent:
            scope.contribute_to_class(view)

    def get_value(self, view):
        """
        Get value used to filter qs_or_objs's field specified for filtering (parent_field in init).
        :param view: DRF view instance - as it provides access to both request and kwargs.
        :return: value to filter by.
        """
        raise NotImplementedError('You need to use ScopeQueryset subclass with get_value implemented.')

    def get_queryset(self, view):
        """
        Obtains parent queryset (init's qs_or_obj) along with any chaining (init's parent) required.
        :param view: DRF view instance.
        :return: queryset instance.
        """
        queryset = self.queryset
        for parent in self.parent:
            queryset = parent.child_queryset(queryset, view)
        return queryset

    def get_object(self, view):
        """
        Caching wrapper around _get_object.
        :param view: DRF view instance.
        :return: object (instance of init's qs_or_obj model except shadowed by subclass).
        """
        if self.get_object_handle:
            obj = view.rest_easy_object_cache.get(self.get_object_handle, None)
            if not obj:
                obj = self._get_object(view)
                view.rest_easy_object_cache[self.get_object_handle] = obj
        else:
            obj = self._get_object(view)
        return obj

    def _get_object(self, view):
        """
        Obtains parent object by which view queryset should be filtered.
        :param view: DRF view instance.
        :return: object (instance of init's qs_or_obj model except shadowed by subclass).
        """
        queryset = self.get_queryset(view)
        queryset = queryset.filter(**{self.parent_field: self.get_value(view)})
        try:
            obj = get_object_or_404(queryset)
        except Http404:
            if self.raise_404:
                raise
            obj = None
        return obj

    def child_queryset(self, queryset, view):
        """
        Performs filtering of the view queryset.
        :param queryset: view queryset instance.
        :param view: view object.
        :return: filtered queryset.
        """
        obj = self.get_object(view)
        if obj is None and not self.allow_none:
            return queryset.none()
        return queryset.filter(**{self.related_field: obj})


class UrlKwargScopeQuerySet(ScopeQuerySet):
    """
    ScopeQuerySet that obtains parent object from url kwargs.
    """

    def __init__(self, *args, **kwargs):
        """
        Adds url_kwarg to :class:`rest_easy.views.ScopeQuerySet` init parameters.

        :param args: same as :class:`rest_easy.views.ScopeQuerySet`.
        :param url_kwarg: name of url field to be obtained from view's kwargs. By default it will be inferred as
         model_name_pk.
        :param kwargs: same as :class:`rest_easy.views.ScopeQuerySet`.
        """
        self.url_kwarg = kwargs.pop('url_kwarg', None)
        super(UrlKwargScopeQuerySet, self).__init__(*args, **kwargs)
        if not self.url_kwarg:
            try:
                self.url_kwarg = '{}_pk'.format(self.queryset.model._meta.model_name)  # pylint: disable=protected-access
            except AttributeError:
                raise RestEasyException('Either related_field or qs_or_obj must be given.')

    def get_value(self, view):
        """
        Obtains value from url kwargs.
        :param view: DRF view instance.
        :return: Value determining parent object.
        """
        return view.kwargs.get(self.url_kwarg)


class RequestAttrScopeQuerySet(ScopeQuerySet):
    """
    ScopeQuerySet that obtains parent object from view's request property.

    It can work two-fold:

    * the request's property contains full object: in this case no filtering of parent's queryset is required. When
      using such approach, is_object must be set to True, and qs_or_obj can be None. Chaining will be disabled since it
      is inherent to filtering process.
    * the request's property contains object's id, uuid, or other unique property. In that case is_object needs to be
      explicitly set to False, and qs_or_obj needs to be a Django model or queryset. Chaining will be performed as
      usually.

    """

    def __init__(self, *args, **kwargs):
        """
        Adds is_object and request_attr  to :class:`rest_easy.views.ScopeQuerySet` init parameters.

        :param args: same as :class:`rest_easy.views.ScopeQuerySet`.
        :param request_attr: name of property to be obtained from view.request.
        :param is_object: if request's property will be an object or a value to filter by. True by default.
        :param kwargs: same as :class:`rest_easy.views.ScopeQuerySet`.
        """
        self.request_attr = kwargs.pop('request_attr', None)
        if self.request_attr is None:
            raise RestEasyException('request_attr must be set explicitly on an {} init.'.format(
                self.__class__.__name__))
        self.is_object = kwargs.pop('is_object', True)
        super(RequestAttrScopeQuerySet, self).__init__(*args, **kwargs)

    def get_value(self, view):
        """
        Obtains value from url kwargs.
        :param view: DRF view instance.
        :return: Value determining parent object.
        """
        return getattr(view.request, self.request_attr, None)

    def _get_object(self, view):
        """
        Extends standard _get_object's behaviour with handling values that are already objects.
        :param view: DRF view instance.
        :return: object to filter view's queryset by.
        """
        if self.is_object:
            return self.get_value(view)
        return super(RequestAttrScopeQuerySet, self)._get_object(view)



# coding: utf-8
# pylint: disable=too-few-public-methods
"""
This module provides redefined DRF's generic views and viewsets leveraging serializer registration.

One of the main issues with creating traditional DRF APIs is a lot of bloat (and we're writing Python, not Java or C#,
to avoid bloat) that's completely unnecessary in a structured Django project. Therefore, this module aims to provide
a better and simpler way to write simple API endpoints - without limiting the ability to create more complex views.
The particular means to that end are:

* :class:`rest_easy.scopes.ScopeQuerySet` and its subclasses (:class:`rest_easy.scopes.UrlKwargScopeQuerySet` and
  :class:`rest_easy.scopes.RequestAttrScopeQuerySet`) provide a simple way to scope views and viewsets.
  by resource (ie. limiting results to single account, or /resource/<resource_pk>/inner_resource/<inner_resource_pk>/)
* generic views leveraging the above, as well as model-and-schema specification instead of queryset, serializer and
  helper methods - all generic views that were available in DRF as well as GenericAPIView are redefined to support
  this.
* Generic :class:`rest_easy.views.ModelViewSet` which allows for very simple definition of resource
  endpoint.

To make the new views work, all that\'s required is a serializer::

    from users.models import User
    from accounts.models import Account
    from rest_easy.serializers import ModelSerializer
    class UserSerializer(ModelSerializer):
        class Meta:
            model = User
            fields = '__all__'
            schema = 'default'

    class UserViewSet(ModelViewSet):
        model = User
        scope = UrlKwargScopeQuerySet(Account)

and in urls.py::

    from django.conf.urls import url, include
    from rest_framework.routers import DefaultRouter
    router = DefaultRouter()
    router.register(r'accounts/(?P<account_pk>[0-9]+)/users', UserViewSet)
    urlpatterns = [url(r'^', include(router.urls))]

The above will provide the users scoped by account primary key as resources: with list, retrieve, create, update and
partial update methods, as well as standard HEAD and OPTIONS autogenerated responses.

You can easily add custom paths to viewsets when needed - it's described in DRF documentation.
"""

from django.conf import settings
from rest_framework.viewsets import ViewSetMixin
from rest_framework import generics, mixins
from six import with_metaclass

from rest_easy.exceptions import RestEasyException
from rest_easy.registers import serializer_register
from rest_easy.scopes import ScopeQuerySet

__all__ = ['GenericAPIView', 'CreateAPIView', 'ListAPIView', 'RetrieveAPIView', 'DestroyAPIView', 'UpdateAPIView',
           'ListCreateAPIView', 'RetrieveUpdateAPIView', 'RetrieveDestroyAPIView', 'RetrieveUpdateDestroyAPIView',
           'ReadOnlyModelViewSet', 'ModelViewSet']


def get_additional_bases():
    """
    Looks for additional view bases in settings.REST_EASY_VIEW_BASES.
    :return:
    """
    resolved_bases = []
    from importlib import import_module
    for base in getattr(settings, 'REST_EASY_VIEW_BASES', []):
        mod, cls = base.rsplit('.', 1)
        resolved_bases.append(getattr(import_module(mod), cls))

    return resolved_bases


def get_additional_mixins():
    """
    Looks for additional view bases in settings.REST_EASY_VIEW_MIXINS.
    :return:
    """
    resolved_bases = []
    from importlib import import_module
    for base in getattr(settings, 'REST_EASY_GENERIC_VIEW_MIXINS', []):
        mod, cls = base.rsplit('.', 1)
        resolved_bases.append(getattr(import_module(mod), cls))

    return resolved_bases

ADDITIONAL_MIXINS = get_additional_mixins()


class ScopedViewMixin(object):
    """
    This class provides a get_queryset method that works with ScopeQuerySet.

    Queryset obtained from superclass is filtered by view.scope's (if it exists) child_queryset() method.
    """

    def get_queryset(self):
        """
        Calls scope's child_queryset methods on queryset as obtained from superclass.
        :return: queryset.
        """
        queryset = super(ScopedViewMixin, self).get_queryset()
        if hasattr(self, 'scope') and self.scope:
            for scope in self.scope:
                queryset = scope.child_queryset(queryset, self)
        return queryset

    def get_scoped_object(self, handle):
        """
        Obtains object from scope when scope's get_object_handle was set.
        :param handle: get_object_handle used in scope initialization.
        :return: object used by scope to filter.
        """
        scope = self.rest_easy_available_object_handles.get(handle, None)
        if scope:
            return scope.get_object(self)
        raise AttributeError('{} get_object handle not found on object {}'.format(handle, self))

    def __getattr__(self, item):
        """
        A shortcut providing get_{get_object_handle} to be able to easily access objects used by this view's scopes
        for filtering. For example, scope = UrlKwargScopeQuerySet(Account) will be available with self.get_account().
        :param item: item to obtain plus 'get_' prefix
        :return: object used by scope for filtering.
        """
        if not item.startswith('get_'):
            raise AttributeError('{} not found on object {}'.format(item, self))
        handle = item[4:]
        try:
            return self.get_scoped_object(handle)
        except AttributeError:
            raise AttributeError('{} not found on object {}'.format(item, self))


class ViewEasyMetaclass(type):  # pylint: disable=too-few-public-methods
    """
    This metaclass sets default queryset on a model-and-schema based views and fills in concrete views with bases.

    It's required for compatibility with some of DRF's elements, like routers.
    """

    def __new__(mcs, name, bases, attrs):
        """
        Create the class.
        """
        if ('queryset' not in attrs or attrs['queryset'] is None) and 'model' in attrs:
            attrs['queryset'] = attrs['model'].objects.all()
        if 'scope' in attrs and isinstance(attrs['scope'], ScopeQuerySet):
            attrs['scope'] = [attrs['scope']]
        attrs['rest_easy_available_object_handles'] = {}
        cls = super(ViewEasyMetaclass, mcs).__new__(mcs, name, bases, attrs)
        for scope in getattr(cls, 'scope', []):
            scope.contribute_to_class(cls)
        return cls


class ChainingCreateUpdateMixin(object):
    """
    Chain-enabled versions of perform_create and perform_update.
    """

    def perform_create(self, serializer, **kwargs):  # pylint: disable=no-self-use
        """
        Extend default implementation with kwarg chaining.
        """
        return serializer.save(**kwargs)

    def perform_update(self, serializer, **kwargs):  # pylint: disable=no-self-use
        """
        Extend default implementation with kwarg chaining.
        """
        return serializer.save(**kwargs)


class GenericAPIViewBase(ScopedViewMixin, generics.GenericAPIView):
    """
    Provides a base for all generic views and viewsets leveraging registered serializers and ScopeQuerySets.

    Adds additional DRF-verb-wise override for obtaining serializer class: serializer_schema_for_verb property.
    It should be a dictionary of DRF verbs and serializer schemas (they work in conjunction with model property).
        serializer_schema_for_verb = {'update': 'schema-mutate', 'create': 'schema-mutate'}
    The priority for obtaining serializer class is:

    * get_serializer_class override
    * serializer_class property
    * model + serializer_schema_for_verb[verb] lookup in :class:`rest_easy.registers.SerializerRegister`
    * model + schema lookup in :class:`rest_easy.registers.SerializerRegister`

    """
    serializer_schema_for_verb = {}

    def __init__(self, **kwargs):
        """
        Set object cache to empty dict.
        :param kwargs: Passthrough to Django view.
        """
        super(GenericAPIViewBase, self).__init__(**kwargs)
        self.rest_easy_object_cache = {}

    def get_drf_verb(self):
        """
        Obtain the DRF verb used for a request.
        """
        method = self.request.method.lower()
        if method == 'get':
            if self.lookup_url_kwarg in self.kwargs:
                return 'retrieve'
        mapping = {
            'get': 'list',
            'post': 'create',
            'put': 'update',
            'patch': 'partial_update',
            'delete': 'destroy'
        }
        return mapping.get(method, method)

    def get_serializer_name(self, verb=None):
        """
        Obtains registered serializer name for this view.

        Leverages :class:`rest_easy.registers.SerializerRegister`. Works when either of or both model
        and schema properties are available on this view.

        :return: registered serializer key.
        """
        model = getattr(self, 'model', None)
        schema = None
        if not model and not hasattr(self, 'schema') and (verb and verb not in self.serializer_schema_for_verb):
            raise RestEasyException('Either model or schema fields need to be set on a model-based GenericAPIView.')
        if verb:
            schema = self.serializer_schema_for_verb.get(verb, None)
        if schema is None:
            schema = getattr(self, 'schema', 'default')
        return serializer_register.get_name(model, schema)

    def get_serializer_class(self):
        """
        Gets serializer appropriate for this view.

        Leverages :class:`rest_easy.registers.SerializerRegister`. Works when either of or both model
        and schema properties are available on this view.

        :return: serializer class.
        """

        if hasattr(self, 'serializer_class') and self.serializer_class:
            return self.serializer_class

        serializer = serializer_register.lookup(self.get_serializer_name(verb=self.get_drf_verb()))
        if serializer:
            return serializer

        raise RestEasyException(u'Serializer for model {} and schema {} cannot be found.'.format(
            getattr(self, 'model', '[no model]'),
            getattr(self, 'schema', '[no schema]')
        ))


class GenericAPIView(with_metaclass(ViewEasyMetaclass, *(get_additional_bases() + [GenericAPIViewBase]))):
    """
    Base view with compat metaclass.
    """
    __abstract__ = True


def create(self, request, *args, **kwargs):  # pragma: no cover
    """
    Shortcut method.
    """
    return self.create(request, *args, **kwargs)


def list_(self, request, *args, **kwargs):  # pragma: no cover
    """
    Shortcut method.
    """
    return self.list(request, *args, **kwargs)


def retrieve(self, request, *args, **kwargs):  # pragma: no cover
    """
    Shortcut method.
    """
    return self.retrieve(request, *args, **kwargs)


def destroy(self, request, *args, **kwargs):  # pragma: no cover
    """
    Shortcut method.
    """
    return self.destroy(request, *args, **kwargs)


def update(self, request, *args, **kwargs):  # pragma: no cover
    """
    Shortcut method.
    """
    return self.update(request, *args, **kwargs)


def partial_update(self, request, *args, **kwargs):  # pragma: no cover
    """
    Shortcut method.
    """
    return self.partial_update(request, *args, **kwargs)


CreateAPIView = type('CreateAPIView',
                     tuple(ADDITIONAL_MIXINS + [ChainingCreateUpdateMixin, mixins.CreateModelMixin, GenericAPIView]),
                     {'post': create,
                      '__doc__': "Concrete view for retrieving or deleting a model instance."})

ListAPIView = type('ListAPIView',
                   tuple(ADDITIONAL_MIXINS + [mixins.ListModelMixin, GenericAPIView]),
                   {'get': list_,
                    '__doc__': "Concrete view for listing a queryset."})


RetrieveAPIView = type('RetrieveAPIView',
                       tuple(ADDITIONAL_MIXINS + [mixins.RetrieveModelMixin, GenericAPIView]),
                       {'get': retrieve,
                        '__doc__': "Concrete view for retrieving a model instance."})


DestroyAPIView = type('DestroyAPIView',
                      tuple(ADDITIONAL_MIXINS + [mixins.DestroyModelMixin, GenericAPIView]),
                      {'delete': destroy,
                       '__doc__': "Concrete view for deleting a model instance."})


UpdateAPIView = type('UpdateAPIView',
                     tuple(ADDITIONAL_MIXINS + [ChainingCreateUpdateMixin, mixins.UpdateModelMixin, GenericAPIView]),
                     {'put': update,
                      'patch': partial_update,
                      '__doc__': "Concrete view for updating a model instance."})


ListCreateAPIView = type('ListCreateAPIView',
                         tuple(ADDITIONAL_MIXINS +
                               [ChainingCreateUpdateMixin, mixins.ListModelMixin,
                                mixins.CreateModelMixin, GenericAPIView]),
                         {'get': list_,
                          'post': create,
                          '__doc__': "Concrete view for listing a queryset or creating a model instance."})


RetrieveUpdateAPIView = type('RetrieveUpdateAPIView',
                             tuple(ADDITIONAL_MIXINS +
                                   [ChainingCreateUpdateMixin, mixins.RetrieveModelMixin,
                                    mixins.UpdateModelMixin, GenericAPIView]),
                             {'get': retrieve,
                              'put': update,
                              'patch': partial_update,
                              '__doc__': "Concrete view for retrieving, updating a model instance."})


RetrieveDestroyAPIView = type('RetrieveDestroyAPIView',
                              tuple(ADDITIONAL_MIXINS +
                                    [mixins.RetrieveModelMixin, mixins.DestroyModelMixin, GenericAPIView]),
                              {'get': retrieve,
                               'delete': destroy,
                               '__doc__': "Concrete view for retrieving or deleting a model instance."})


RetrieveUpdateDestroyAPIView = type('RetrieveUpdateDestroyAPIView',
                                    tuple(ADDITIONAL_MIXINS +
                                          [ChainingCreateUpdateMixin, mixins.RetrieveModelMixin,
                                           mixins.UpdateModelMixin, mixins.DestroyModelMixin, GenericAPIView]),
                                    {'get': retrieve,
                                     'put': update,
                                     'patch': partial_update,
                                     'delete': destroy,
                                     '__doc__': "Concrete view for retrieving, updating or deleting a model instance."
                                    })


class GenericViewSet(ViewSetMixin, GenericAPIView):  # pragma: no cover
    """
    The GenericViewSet class does not provide any actions by default,
    but does include the base set of generic view behavior, such as
    the `get_object` and `get_queryset` methods.
    """


ReadOnlyModelViewSet = type('ReadOnlyModelViewSet',
                            tuple(ADDITIONAL_MIXINS +
                                  [mixins.RetrieveModelMixin, mixins.ListModelMixin, GenericViewSet]),
                            {'__doc__': "A viewset that provides default `list()` and `retrieve()` actions."})


ModelViewSet = type('ModelViewSet',
                    tuple(ADDITIONAL_MIXINS +
                          [ChainingCreateUpdateMixin, mixins.CreateModelMixin, mixins.RetrieveModelMixin,
                           mixins.UpdateModelMixin, mixins.DestroyModelMixin, mixins.ListModelMixin, GenericViewSet]),
                    {'__doc__': "A viewset that provides default `create()`, `retrieve()`, `update()`, "
                                "`partial_update()`, `destroy()` and `list()` actions."})
